/* -------------------------------------------------------------------------------------------------------------------------
* Project           :C-to-FPGA Solution for Acclerating Tomographic Reconstruction
* File Name         :uwer_fbp.c (from project UW Emission Reconstruction) 
*                    Copyright (c) 2004 University of Washington - Imaging Research Laboratory. All rights reserved.
* Version           :2.0 (V2)
* (V1)Author        :Adam Alessio ( * Modified from code written by Charles Bouman and Ken Sauer.)
* (V1)Created       :Wed, Jan 14 2004
* (V2)Author        :Nikhil Subramanian
* (V2)Created       :Wed, Jan 14 2009
* References        :
* Description       :Performs FBP. V2 adds function call to impulse C backproject function
*
******This file may not be distributed or modified without this header*****
*-------------------------------------------------------------------------------------------------------------------------*/



#include "uwer_fbp.h"
#include "mem_param.h"


/*Global variable and functions declared here*/
float **sinogram_to_send = NULL;
extern float image_hw[IMG_SIZE][IMG_SIZE];
extern void run_hw_processes();


int VERBOSE=1; 

void fbp_recon(struct sino_info *sinoInfo,
		  struct recon_image_info *imageInfo,
	       struct exec_parameters *params)
{
  double totimg, avgimg;    //To normalize the image 
  int i,j;
  int frame, plane;
  double pmass;
  float **sinogramData, **image;
  struct fbp_filter filter;
  
  Print("-----Performing Conventional FBP\n");

  filter.type = params->fbp_type;
  filter.cutoff = params->fbp_cutoff;

  apply_attenuation_correction(params, sinoInfo);
  
  for (frame=0;frame<imageInfo->nFrames;frame++)
    for (plane=params->ReconPlanes[0];plane<=params->ReconPlanes[1];plane++)
      {
	Print("Reconstructing Frame:%d of %d and Plane:%d of [%d,%d]\n", frame, imageInfo->nFrames,plane,params->ReconPlanes[0],params->ReconPlanes[1]);

	sinogramData = sinoInfo->sinogramData[frame][plane];
	image = imageInfo->imageData[frame][plane];
	sinogram_to_send = sinogramData;
	/* Calculate pmass of input sinogram for mean */
	/* value normalization later                  */
	pmass=calc_pmass(sinogramData, sinoInfo, imageInfo);

	/* Filter Sinogram */
	filter_sino(sinogramData,sinoInfo,&filter);
	
	/* Backprojection */
    printf("\n\n\nStarting Back Projection in Hardware\n\n\n");
	backproject(image,sinogramData,sinoInfo,imageInfo);
    printf("\n\n\nHardware is Done\n\n\n");
	
	/* mean-value correction */
	mean_value_correct(image, pmass, imageInfo);
    
	fflush(stdout);fflush(stderr);
      }
  
  
}       


void filter_sino(float **sinogramData,struct sino_info *sino,struct fbp_filter* filter)
{
  double *data,*ans;
  int n,i,j;

  /* printf("Filtering...\n"); */

  /* Calculate filter-length */
  n = (int) ceil(log(sino->nProjs)/log(2.0));
  n = pow(2,n);

  /* Calculate Filter  of length (2*n) */
  if (filter->type == RAMP_FILTER)
    Ramp_Filter(n,sino,filter);
  else if (filter->type == GHAMMING_FILTER)
    GHamming_Filter(n,sino,filter);

 /*  normalize_filter_to_sum1(filter->filter, 2*n);  *//* AMA Added 10/11/02 */
      /* Don't need, the height set with proj->spacing in the filter creation results in correctly scaled image...*/

  /* Allocate space for data */
  ans = (double *)get_spc(4*n, sizeof(double));
  data = (double *)get_spc(2*n, sizeof(double));

  /* Pad data with zeros */
  for(i=sino->nProjs;i<2*n;i++)
    data[i] = 0.0;
 

  /* Filter sinogram  for each projection angle */
  for(i=0;i<sino->nAngles;i++) 
    {
      for(j=0;j<sino->nProjs;j++)
	data[j] = (double)sinogramData[i][j];
  
      convlv(data-1,2*n,(filter->filter)-1,ans-1);
      
      /* insert into Y */
      for(j=0;j<sino->nProjs;j++)
	sinogramData[i][j] = (float)ans[j]; 
      
    }


/* #define WRITE_FILTER */
#ifdef WRITE_FILTER
  {
    FILE* fp;

    /* Write filter response to disk for checking */
    fp=fopen("filter","wa");
    for (i=0;i<2*n;i++)
      fprintf(fp,"%g ",filter->filter[i]);
    fclose(fp);
  }
  {
    FILE* fp;
   fp=fopen("sino.yes","wa");
   for(i=0;i<sino->nAngles;i++) {
     for(j=0;j<sino->nProjs;j++)
       fprintf(fp,"%g ",sinogramData[i][j]);
     fprintf(fp,"\n");
   }
    fclose(fp);
  }

#endif
  
  free((void *)(ans));
  free((void *)(data));
  free((void *)(filter->filter));
} 




/* This reconstruction filter GHamming_Filter has frequency response

      H(omega) = H_ramp(omega) * (0.5 + 0.5cos(pi*omega/omega_c))

   where H_ramp is the ideal ramp reconstruction filter, 
   and omega_c is the cutoff frequency. 
   The filter input parameter specifies the ratio omega/omega_c .
*/
void GHamming_Filter(int n,struct sino_info *sino,struct fbp_filter* filter)
{
  int i;
  double del;

  /* allocate filter */
  filter->filter = (double *)get_spc(2*n, sizeof(double));
  
  del = filter->cutoff;
  /* printf("cos delay=%f, so cutoff is %f percent.\n",del,100.0*del);  */
  
  /* Calculate impulse response in time domain, 
     hrl is ideal ramp filter response 1/(4*projSpacing)*(2*delta(n)-sinc^2(n/2)),
     dhrl is basically same as hrl but w/ double arithm. input, 
          and uses 1/(4*projSpacing)*(2*sinc(n)-sinc^2(n/2))
  */
  for(i=-n+1;i<=n;i++)
    filter->filter[i+n-1] = 0.5*(hrl(i,sino) + 0.5*(dhrl(i-1.0/del,sino) + dhrl(i+1.0/del,sino) )); 
  
  /* So far, filter response is only accurate for omega<=omega_c */
  /* convolve with sinc() to chop off frequency domain portion
     that is over cutoff frequency */
  cut_freq(filter->filter,del,n);
  

  /* Resort filter : before, filter had phase delay,*/
  /* i.e. ideal h[0] was at filter[len-1], now for fft-filtering */
  /* resort to fft-symmetry to get lin. phase */
 
  for (i=0;i<=n;i++)
    filter->filter[i]=filter->filter[i+n-1];
  
  for (i=n+1;i<2*n;i++)
    filter->filter[i]=filter->filter[2*n-i];
  
}


/* This routine implements the ideal ramp filter but with 
   a cutoff frequency as specified on the commandline. 
   So
   H(omega) = H_ramp(omega) * rect(omega/omega_c) 

   where rect(x)=1 for |x|<1
   and H_ramp is the ideal highpass.
*/
void Ramp_Filter(int n,struct sino_info *sino,struct fbp_filter* filter)
{
  int i;
  double del;

  /* allocate filter */
  filter->filter = (double *)get_spc(2*n, sizeof(double));
  
  del = filter->cutoff;
  printf("cos delay=%f, so cutoff is %f percent.\n",del,100.0*del); 
  
  /* Calculate impulse response in time domain, 
     hrl is ideal ramp filter response 1/(4*projSpacing)*(2*delta(n)-sinc^2(n/2)),
   */
  for(i=-n+1;i<=n;i++)
    filter->filter[i+n-1] = hrl(i,sino); 
  
  /* So far, filter response is only accurate for omega<=omega_c */
  /* convolve with sinc() to chop off frequency domain portion
     that is over cutoff frequency */
  cut_freq(filter->filter,del,n);
  

  /* Resort filter : before, filter had phase delay,*/
  /* i.e. ideal h[0] was at filter[len-1], now for fft-filtering */
  /* resort to fft-symmetry to get lin. phase */
 
  for (i=0;i<=n;i++)
    filter->filter[i]=filter->filter[i+n-1];
  
  for (i=n+1;i<2*n;i++)
    filter->filter[i]=filter->filter[2*n-i];

  printf("Finished Ramp\n");
}




/* convolution , used in compute_filter to */
/* convolve Gaussian*HammW with ideal recon-filter*/

void conv(double *out, double *in1, int len1, double *in2, int len2)
{
  int i,n;

  for (n=0;n<len1+len2-1;n++)
    {
      out[n]=0.0;
      for (i=0;i<len1;i++)
	if ((n-i>=0) && (n-i<len2))
	  out[n] += (in1[i]*in2[n-i]);
    }
}



void normalize_filter_to_sum1(double *filt, int len)
{
  int i;
  double sum=0.0;
  FILE *fp;

/*   fp = fopen("AMAFilt.dat","w"); */
  
/*   for (i=0;i<len;i++) */
/*     fprintf(fp,"%f\n",filt[i]); */

/*   fclose(fp); */

  for (i=0;i<len;i++)
    sum+=(filt[i]);

  // printf("sum = %f\n",sum); 

  for (i=0;i<len;i++)
    filt[i]/=sum;
}




/* hrl returns 1/(4*projSpacing)*(   2*delta(m)- (sinc(m/2))^2)  */
/* this is the ideal reconstruction filter */

double hrl(int m, struct sino_info *sino)
{ 
    float temp;

    if (m == 0)
        return(1/(4.0*sino->projSpacing));
    else
    {
        temp = M_PI*m/2;
        temp = sin(temp)/temp;
        temp =  - temp*temp;
        return( temp/(4.0*(double)sino->projSpacing) );
    }
}

double dhrl(double m, struct sino_info *sino)
{ 
    double temp,temp2,omeg;

    if (fabs(m)<1e-8)
        return(1.0/(4.0*sino->projSpacing));
    else
    {
        omeg = M_PI*m/2.0;
        temp = sin(omeg)/omeg;
        temp2 = 2*sin(omeg*2)/(omeg*2);
        temp = temp2 -temp*temp;
        return( temp/(4.0*(double)sino->projSpacing) );
    }
}




void backproject(float **image,float **sinogramData,struct sino_info *sino,struct recon_image_info *geom)
{ 
  
  int i,j;  	
  
  run_hw_processes();

  for(i=0;i<IMG_SIZE;++i)
      for(j=0;j<IMG_SIZE;++j)
			image[i][j]=image_hw[i][j];
}



/* Interpolate is used in backproject to */
/* perform bi-linear interploation between parallel projections */

double interpolate(double findex, float **Y, int angle_index)
{
  int index,ce,fl;
  double out;
  
  /* check whether findex is integer */
  index=(int)findex;
  if ((findex-(double)index>-1E-10) && (findex-(double)index<1E-10))
    out=(double)Y[angle_index][index];
  else /* if not integer, then bilinear interpolation between parallel projections */
    {
      ce=(int)(findex+1.0); /* ceiling */
      fl=(int)findex; /* floor  */

      out=(((double)ce)-findex)*(double)Y[angle_index][fl] + (findex-(double)fl)*(double)Y[angle_index][ce];
    }
 
  return(out);

}


void mean_value_correct(float **image, double pmass, struct recon_image_info *geom)
{
  double imass=0.0;
  double  cor_factor,mx;
  int    i,j;
  
  for (i=0;i<geom->yDim;i++)
    for (j=0;j<geom->xDim;j++)
      imass += (double)image[i][j];
  // imass /= ((double)(geom->yDim*geom->xDim)); 
  
  cor_factor=(pmass/imass);
  
  printf("pmass:%f imass:%f cor_factor:%f\n",pmass,imass,cor_factor); 

  for (i=0;i<geom->yDim;i++)
    for (j=0;j<geom->xDim;j++)
      image[i][j] *= cor_factor;

  /* Calculate Mean just to printf */
  mx = 0.0;
  for (i=0;i<geom->yDim;i++)
    for (j=0;j<geom->xDim;j++)
      mx += image[i][j];

  mx /= (double)(geom->yDim*geom->xDim);

 
  printf("The mean of the CBP image is %f\n",mx);
  printf("The min,max of the CBP image are %.3f,%.3f\n",min(image,geom->yDim,geom->xDim),max(image,geom->yDim,geom->xDim)); 
} 




/* Calculates sinogram mean and re-normalizes it     */
/* for the reconstruction geometry in geom           */
/* The output is later used in mean_value_correction */

double calc_pmass(float **sinogramData, struct sino_info *sino, struct recon_image_info *imageInfo)
{
  double pmass;
  double mean=0.0;
  int i,j;
  
  for (i=0;i<sino->nAngles;i++)
    for (j=0;j<sino->nProjs;j++)
      mean +=(double)sinogramData[i][j];

  //  mean = mean/((double)(sino->nAngles*sino->nProjs));
  // printf("early pmass = %f\n", mean);
  //  pmass = mean*((double)sino->nProjs)*sino->projSpacing/(imageInfo->xPixelSize*imageInfo->yPixelSize*(double)(imageInfo->yDim*imageInfo->xDim)); 

  pmass = mean;
  return(pmass);
}



double min(float **img,int N, int M)
{
  int i,j;
  double min;
  
  min = img[0][0];
  for (i=0;i<N;i++)
    for(j=0;j<M;j++)
      if ( img[i][j] < min )
	min = img[i][j];
  
  return(min);
}


double max(float **img,int N, int M)
{
  int i,j;
  double max;
  
  max = img[0][0];
  
  for (i=0;i<N;i++)
    for(j=0;j<M;j++)
      if (img[i][j] > max)
	max = img[i][j];
  
  return(max);
}




void cut_freq(double *filter,double del,int n)
{
  double flt2[2048];
  int i,k;
  
  for(i=-n+1;i<=n;i++)
    {
      flt2[i+n-1]=0.0;
      for (k=-n+1;k<=n;k++)
	flt2[i+n-1] += (filter[k+n-1] * snc(del*(double)(i-k))*del);
    } 
  for(i=-n+1;i<=n;i++)
    filter[i+n-1] = flt2[i+n-1];
  
}

double snc(double n)
{
  double out,x;

  if (fabs(n)<1e-9)
    out=1.0;
  else
    {
      x=M_PI*n;
      out=sin(x) / x;
    }
  return(out);


}




